@@ -32,6 +32,10 @@
 #include "log_data_helper.hh"
 
 #include "config.h"
+#include "logfile.hh"
+#include "pugixml/pugixml.hpp"
+#include "sql_util.hh"
+#include "xml_util.hh"
 
 void
 log_data_helper::clear()
@@ -50,13 +54,12 @@ log_data_helper::clear()
 bool
 log_data_helper::parse_line(content_line_t line, bool allow_middle)
 {
-    logfile::iterator ll;
-    bool retval = false;
+    auto retval = false;
 
     this->ldh_source_line = this->ldh_line_index = line;
 
     this->ldh_file = this->ldh_log_source.find(this->ldh_line_index);
-    ll = this->ldh_file->begin() + this->ldh_line_index;
+    auto ll = this->ldh_file->begin() + this->ldh_line_index;
     this->ldh_y_offset = 0;
     while (allow_middle && ll->is_continued()) {
         --ll;
@@ -64,7 +67,6 @@ log_data_helper::parse_line(content_line_t line, bool allow_middle)
     }
     this->ldh_line = ll;
     if (!ll->is_message()) {
-        log_warning("failed to parse line %d", line);
         this->ldh_parser.reset();
         this->ldh_scanner.reset();
         this->ldh_namer.reset();
@@ -74,14 +76,17 @@ log_data_helper::parse_line(content_line_t line, bool allow_middle)
         this->ldh_line_attrs.clear();
     } else {
         auto format = this->ldh_file->get_format();
-        struct line_range body;
+        line_range body;
         auto& sa = this->ldh_line_attrs;
 
         this->ldh_line_attrs.clear();
         this->ldh_line_values.clear();
         this->ldh_file->read_full_message(ll, this->ldh_line_values.lvv_sbr);
         this->ldh_line_values.lvv_sbr.erase_ansi();
-        format->annotate(this->ldh_line_index, sa, this->ldh_line_values);
+        format->annotate(this->ldh_file.get(),
+                         this->ldh_line_index,
+                         sa,
+                         this->ldh_line_values);
 
         body = find_string_attr_range(sa, &SA_BODY);
         if (body.lr_start == -1) {
@@ -95,7 +100,9 @@ log_data_helper::parse_line(content_line_t line, bool allow_middle)
             = std::make_unique<data_parser>(this->ldh_scanner.get());
         this->ldh_msg_format.clear();
         this->ldh_parser->dp_msg_format = &this->ldh_msg_format;
-        this->ldh_parser->parse();
+        if (body.length() < 128 * 1024) {
+            this->ldh_parser->parse();
+        }
         this->ldh_namer
             = std::make_unique<column_namer>(column_namer::language::SQL);
         this->ldh_extra_json.clear();
@@ -114,7 +121,9 @@ log_data_helper::parse_line(content_line_t line, bool allow_middle)
             if (ldh_line_value.lv_meta.lvm_column
                     .is<logline_value_meta::external_column>())
             {
-                char buf[ldh_line_value.lv_meta.lvm_name.size() + 2];
+                stack_buf allocator;
+                auto* buf = allocator.allocate(
+                    ldh_line_value.lv_meta.lvm_name.size() + 2);
 
                 auto rc = fmt::format_to(
                     buf, FMT_STRING("/{}"), ldh_line_value.lv_meta.lvm_name);
