@@ -31,7 +31,8 @@
 
 #include "base/itertools.hh"
 #include "config.h"
-#include "scn/scn.h"
+#include "document.sections.hh"
+#include "scn/scan.h"
 
 static std::vector<plain_text_source::text_line>
 to_text_line(const std::vector<attr_line_t>& lines)
@@ -49,17 +50,13 @@ to_text_line(const std::vector<attr_line_t>& lines)
            });
 }
 
-plain_text_source::plain_text_source(const std::string& text)
+plain_text_source::plain_text_source(const string_fragment& text)
 {
-    size_t start = 0, end;
+    size_t start = 0;
 
-    while ((end = text.find('\n', start)) != std::string::npos) {
-        size_t len = (end - start);
-        this->tds_lines.emplace_back(start, text.substr(start, len));
-        start = end + 1;
-    }
-    if (start < text.length()) {
-        this->tds_lines.emplace_back(start, text.substr(start));
+    for (const auto& line : text.split_lines()) {
+        this->tds_lines.emplace_back(start, attr_line_t::from_ansi_frag(line));
+        start += line.length();
     }
     this->tds_longest_line = this->compute_longest_line();
 }
@@ -86,7 +83,7 @@ plain_text_source::replace_with(const attr_line_t& text_lines)
         lines.pop_back();
     }
     for (auto& line : lines) {
-        auto line_len = line.length() + 1;
+        auto line_len = line.al_string.size() + 1;
         this->tds_lines.emplace_back(off, std::move(line));
         off += line_len;
     }
@@ -101,9 +98,10 @@ plain_text_source&
 plain_text_source::replace_with_mutable(attr_line_t& text_lines,
                                         text_format_t tf)
 {
+    this->tds_text_format = tf;
     this->tds_lines.clear();
     this->tds_doc_sections
-        = lnav::document::discover_structure(text_lines, line_range{0, -1}, tf);
+        = lnav::document::discover(text_lines).with_text_format(tf).perform();
     file_off_t off = 0;
     auto lines = text_lines.split_lines();
     while (!lines.empty() && lines.back().empty()) {
@@ -126,7 +124,7 @@ plain_text_source::replace_with(const std::vector<std::string>& text_lines)
 {
     file_off_t off = 0;
     for (const auto& str : text_lines) {
-        this->tds_lines.emplace_back(off, str);
+        this->tds_lines.emplace_back(off, attr_line_t::from_ansi_str(str));
         off += str.length() + 1;
     }
     this->tds_longest_line = this->compute_longest_line();
@@ -140,6 +138,7 @@ plain_text_source&
 plain_text_source::replace_with(const std::vector<attr_line_t>& text_lines)
 {
     file_off_t off = 0;
+    this->tds_lines.clear();
     for (const auto& al : text_lines) {
         this->tds_lines.emplace_back(off, al);
         off += al.length() + 1;
@@ -180,7 +179,7 @@ plain_text_source::text_line_width(textview_curses& curses)
     return this->tds_longest_line;
 }
 
-void
+line_info
 plain_text_source::text_value_for_line(textview_curses& tc,
                                        int row,
                                        std::string& value_out,
@@ -199,6 +198,8 @@ plain_text_source::text_value_for_line(textview_curses& tc,
             break;
         }
     }
+
+    return {};
 }
 
 void
@@ -211,7 +212,7 @@ plain_text_source::text_attrs_for_line(textview_curses& tc,
         && tc.get_selection() == line)
     {
         value_out.emplace_back(line_range{0, -1},
-                               VC_STYLE.value(text_attrs{A_REVERSE}));
+                               VC_STYLE.value(text_attrs::with_reverse()));
     }
     for (const auto& indent : this->tds_doc_sections.m_indents) {
         if (indent < this->tds_line_indent_size) {
@@ -251,41 +252,45 @@ plain_text_source::compute_longest_line()
     return retval;
 }
 
-nonstd::optional<vis_line_t>
+std::optional<vis_line_t>
 plain_text_source::line_for_offset(file_off_t off) const
 {
     struct cmper {
-        bool operator()(const file_off_t& lhs, const text_line& rhs)
+        bool operator()(const file_off_t& lhs, const text_line& rhs) const
         {
             return lhs < rhs.tl_offset;
         }
 
-        bool operator()(const text_line& lhs, const file_off_t& rhs)
+        bool operator()(const text_line& lhs, const file_off_t& rhs) const
         {
             return lhs.tl_offset < rhs;
         }
     };
 
     if (this->tds_lines.empty()) {
-        return nonstd::nullopt;
+        return std::nullopt;
     }
 
+    log_trace("line_for_offset(%d)", off);
     auto iter = std::lower_bound(
         this->tds_lines.begin(), this->tds_lines.end(), off, cmper{});
     if (iter == this->tds_lines.end()) {
         if (this->tds_lines.back().contains_offset(off)) {
-            return nonstd::make_optional(
+            return std::make_optional(
                 vis_line_t(std::distance(this->tds_lines.end() - 1, iter)));
         }
-        return nonstd::nullopt;
+        return std::nullopt;
     }
 
     if (!iter->contains_offset(off) && iter != this->tds_lines.begin()) {
+        log_trace("  lower_bound (%lld) does not contain offset",
+                  iter->tl_offset);
         --iter;
     }
 
-    return nonstd::make_optional(
-        vis_line_t(std::distance(this->tds_lines.begin(), iter)));
+    auto retval = vis_line_t(std::distance(this->tds_lines.begin(), iter));
+    log_trace("  retval=%d", retval);
+    return retval;
 }
 
 void
@@ -297,12 +302,13 @@ plain_text_source::text_crumbs_for_line(int line,
 
     this->tds_doc_sections.m_sections_tree.visit_overlapping(
         tl.tl_offset,
+        tl.tl_offset + tl.tl_value.length(),
         [&crumbs, initial_size, meta = &this->tds_doc_sections, this](
             const auto& iv) {
             auto path = crumbs | lnav::itertools::skip(initial_size)
                 | lnav::itertools::map(&breadcrumb::crumb::c_key)
                 | lnav::itertools::append(iv.value);
-            crumbs.template emplace_back(
+            crumbs.emplace_back(
                 iv.value,
                 [meta, path]() { return meta->possibility_provider(path); },
                 [this, meta, path](const auto& key) {
@@ -316,7 +322,7 @@ plain_text_source::text_crumbs_for_line(int line,
                     if (parent_node == nullptr) {
                         return;
                     }
-                    key.template match(
+                    key.match(
                         [this, parent_node](const std::string& str) {
                             auto sib_iter
                                 = parent_node->hn_named_children.find(str);
@@ -359,13 +365,13 @@ plain_text_source::text_crumbs_for_line(int line,
         auto poss_provider = [curr_node = node.value()]() {
             std::vector<breadcrumb::possibility> retval;
             for (const auto& child : curr_node->hn_named_children) {
-                retval.template emplace_back(child.first);
+                retval.emplace_back(child.first);
             }
             return retval;
         };
         auto path_performer = [this, curr_node = node.value()](
                                   const breadcrumb::crumb::key_t& value) {
-            value.template match(
+            value.match(
                 [this, curr_node](const std::string& str) {
                     auto child_iter = curr_node->hn_named_children.find(str);
                     if (child_iter != curr_node->hn_named_children.end()) {
@@ -391,10 +397,10 @@ plain_text_source::text_crumbs_for_line(int line,
     }
 }
 
-nonstd::optional<vis_line_t>
+std::optional<vis_line_t>
 plain_text_source::row_for_anchor(const std::string& id)
 {
-    nonstd::optional<vis_line_t> retval;
+    std::optional<vis_line_t> retval;
 
     if (this->tds_doc_sections.m_sections_root == nullptr) {
         return retval;
@@ -411,10 +417,12 @@ plain_text_source::row_for_anchor(const std::string& id)
             auto comp_pair = hier_sf.split_when(string_fragment::tag1{'/'});
             auto scan_res
                 = scn::scan_value<int64_t>(comp_pair.first.to_string_view());
-            if (scan_res && scan_res.empty()) {
-                path.emplace_back(scan_res.value());
+            if (scan_res && scan_res->range().empty()) {
+                path.emplace_back(scan_res->value());
             } else {
-                path.emplace_back(json_ptr::decode(comp_pair.first));
+                stack_buf allocator;
+                path.emplace_back(
+                    json_ptr::decode(comp_pair.first, allocator).to_string());
             }
             hier_sf = comp_pair.second;
         }
@@ -432,8 +440,7 @@ plain_text_source::row_for_anchor(const std::string& id)
         meta.m_sections_root.get(),
         [this, &id, &retval](const lnav::document::hier_node* node) {
             for (const auto& child_pair : node->hn_named_children) {
-                const auto& child_anchor
-                    = text_anchors::to_anchor_string(child_pair.first);
+                const auto& child_anchor = to_anchor_string(child_pair.first);
 
                 if (child_anchor != id) {
                     continue;
@@ -456,20 +463,19 @@ plain_text_source::get_anchors()
         this->tds_doc_sections.m_sections_root.get(),
         [&retval](const lnav::document::hier_node* node) {
             for (const auto& child_pair : node->hn_named_children) {
-                retval.emplace(
-                    text_anchors::to_anchor_string(child_pair.first));
+                retval.emplace(to_anchor_string(child_pair.first));
             }
         });
 
     return retval;
 }
 
-nonstd::optional<std::string>
+std::optional<std::string>
 plain_text_source::anchor_for_row(vis_line_t vl)
 {
-    nonstd::optional<std::string> retval;
+    std::optional<std::string> retval;
 
-    if (vl > this->tds_lines.size()
+    if (vl > (ssize_t) this->tds_lines.size()
         || this->tds_doc_sections.m_sections_root == nullptr)
     {
         return retval;
@@ -481,69 +487,73 @@ plain_text_source::anchor_for_row(vis_line_t vl)
         tl.tl_offset, tl.tl_offset + tl.tl_value.al_string.length());
 
     if (path_for_line.empty()) {
-        return nonstd::nullopt;
+        return std::nullopt;
     }
 
     if ((path_for_line.size() == 1
          || this->tds_text_format == text_format_t::TF_MARKDOWN)
         && path_for_line.back().is<std::string>())
     {
-        return text_anchors::to_anchor_string(
-            path_for_line.back().get<std::string>());
+        return to_anchor_string(path_for_line.back().get<std::string>());
     }
 
-    auto comps = path_for_line | lnav::itertools::map([](const auto& elem) {
-                     return elem.match(
-                         [](const std::string& str) {
-                             return json_ptr::encode_str(str);
-                         },
-                         [](size_t index) { return fmt::to_string(index); });
-                 });
+    auto comps
+        = path_for_line | lnav::itertools::map([](const auto& elem) {
+              return elem.match(
+                  [](const std::string& str) {
+                      stack_buf allocator;
+                      return json_ptr::encode(str, allocator).to_string();
+                  },
+                  [](size_t index) { return fmt::to_string(index); });
+          });
 
     return fmt::format(FMT_STRING("#/{}"),
                        fmt::join(comps.begin(), comps.end(), "/"));
 }
 
-nonstd::optional<vis_line_t>
-plain_text_source::adjacent_anchor(vis_line_t vl, text_anchors::direction dir)
+std::optional<vis_line_t>
+plain_text_source::adjacent_anchor(vis_line_t vl, direction dir)
 {
-    if (vl > this->tds_lines.size()
+    if (vl > (ssize_t) this->tds_lines.size()
         || this->tds_doc_sections.m_sections_root == nullptr)
     {
-        return nonstd::nullopt;
+        return std::nullopt;
     }
 
     const auto& tl = this->tds_lines[vl];
     auto path_for_line = this->tds_doc_sections.path_for_range(
         tl.tl_offset, tl.tl_offset + tl.tl_value.al_string.length());
 
-    auto& md = this->tds_doc_sections;
+    log_trace("adjacent_anchor: curr path = %s",
+              fmt::format(FMT_STRING("{}"), path_for_line).c_str());
+    const auto& md = this->tds_doc_sections;
     if (path_for_line.empty()) {
         auto neighbors_res = md.m_sections_root->line_neighbors(vl);
         if (!neighbors_res) {
-            return nonstd::nullopt;
+            return std::nullopt;
         }
 
         switch (dir) {
-            case text_anchors::direction::prev: {
+            case direction::prev: {
                 if (neighbors_res->cnr_previous) {
                     return this->line_for_offset(
                         neighbors_res->cnr_previous.value()->hn_start);
                 }
                 break;
             }
-            case text_anchors::direction::next: {
+            case direction::next: {
                 if (neighbors_res->cnr_next) {
                     return this->line_for_offset(
                         neighbors_res->cnr_next.value()->hn_start);
-                } else if (!md.m_sections_root->hn_children.empty()) {
+                }
+                if (!md.m_sections_root->hn_children.empty()) {
                     return this->line_for_offset(
                         md.m_sections_root->hn_children[0]->hn_start);
                 }
                 break;
             }
         }
-        return nonstd::nullopt;
+        return std::nullopt;
     }
 
     auto last_key = path_for_line.back();
@@ -552,26 +562,29 @@ plain_text_source::adjacent_anchor(vis_line_t vl, text_anchors::direction dir)
     auto parent_opt = lnav::document::hier_node::lookup_path(
         md.m_sections_root.get(), path_for_line);
     if (!parent_opt) {
-        return nonstd::nullopt;
+        log_trace("  no parent");
+        return std::nullopt;
     }
-    auto parent = parent_opt.value();
+    const auto* parent = parent_opt.value();
 
     auto child_hn = parent->lookup_child(last_key);
     if (!child_hn) {
         // XXX "should not happen"
-        return nonstd::nullopt;
+        return std::nullopt;
     }
 
     auto neighbors_res = parent->child_neighbors(
         child_hn.value(), tl.tl_offset + tl.tl_value.al_string.length() + 1);
     if (!neighbors_res) {
-        return nonstd::nullopt;
+        log_trace("no neighbors");
+        return std::nullopt;
     }
 
     if (neighbors_res->cnr_previous && last_key.is<std::string>()) {
         auto neighbor_sub
             = neighbors_res->cnr_previous.value()->lookup_child(last_key);
         if (neighbor_sub) {
+            log_trace("  loading previous child");
             neighbors_res->cnr_previous = neighbor_sub;
         }
     }
@@ -580,20 +593,23 @@ plain_text_source::adjacent_anchor(vis_line_t vl, text_anchors::direction dir)
         auto neighbor_sub
             = neighbors_res->cnr_next.value()->lookup_child(last_key);
         if (neighbor_sub) {
+            log_trace("  loading next child");
             neighbors_res->cnr_next = neighbor_sub;
         }
     }
 
     switch (dir) {
-        case text_anchors::direction::prev: {
+        case direction::prev: {
             if (neighbors_res->cnr_previous) {
                 return this->line_for_offset(
                     neighbors_res->cnr_previous.value()->hn_start);
             }
             break;
         }
-        case text_anchors::direction::next: {
+        case direction::next: {
             if (neighbors_res->cnr_next) {
+                log_trace("  next offset %d",
+                          neighbors_res->cnr_next.value()->hn_start);
                 return this->line_for_offset(
                     neighbors_res->cnr_next.value()->hn_start);
             }
@@ -601,5 +617,5 @@ plain_text_source::adjacent_anchor(vis_line_t vl, text_anchors::direction dir)
         }
     }
 
-    return nonstd::nullopt;
+    return std::nullopt;
 }
