@@ -154,7 +154,7 @@ insert_sql_help(help_text& root, const help_text& curr)
             insert_sql_help(root, param);
         }
         for (const auto& eval : param.ht_enum_values) {
-            sqlite_function_help.emplace(eval, &root);
+            sqlite_function_help.emplace(eval.to_string(), &root);
         }
     }
 }
@@ -1459,7 +1459,7 @@ register_sqlite_funcs(sqlite3* db, sqlite_registration_func_t* reg_funcs)
             .sql_keyword()
             .with_parameter(help_text("filter", "Additional processing of rows")
                                 .optional()
-                                .with_enum_values({"DISTINCT", "ALL"}))
+                                .with_enum_values({"DISTINCT"_frag, "ALL"_frag}))
             .with_parameter(
                 help_text(
                     "result-column",
@@ -1611,10 +1611,10 @@ register_sqlite_funcs(sqlite3* db, sqlite_registration_func_t* reg_funcs)
             .sql_infix()
             .with_parameter(help_text("nullness")
                                 .with_enum_values({
-                                    "ISNULL",
-                                    "NOTNULL",
-                                    "NOT NULL",
-                                    "IS NOT NULL",
+                                    "ISNULL"_frag,
+                                    "NOTNULL"_frag,
+                                    "NOT NULL"_frag,
+                                    "IS NOT NULL"_frag,
                                 })
                                 .optional())
             .with_example({
@@ -1670,11 +1670,11 @@ register_sqlite_funcs(sqlite3* db, sqlite_registration_func_t* reg_funcs)
             .with_parameter(
                 help_text("direction", "The direction, ASCending or DESCending")
                     .optional()
-                    .with_enum_values({"ASC", "DESC"}))
+                    .with_enum_values({"ASC"_frag, "DESC"_frag}))
             .with_parameter(
                 help_text("null-handling")
                     .optional()
-                    .with_enum_values({"NULLS FIRST", "NULLS LAST"})),
+                    .with_enum_values({"NULLS FIRST"_frag, "NULLS LAST"_frag})),
 
         help_text("select-stmt")
             .with_summary(
