@@ -7,17 +7,15 @@
  * commercial or non-commercial, and by any means.
  */
 
-#ifdef __CYGWIN__
-#    include <alloca.h>
-#endif
-
 #include <unordered_map>
 
 #include <sqlite3.h>
 #include <stdlib.h>
 #include <string.h>
 
+#include "base/fts_fuzzy_match.hh"
 #include "base/humanize.hh"
+#include "base/is_utf8.hh"
 #include "base/lnav.gzip.hh"
 #include "base/string_util.hh"
 #include "column_namer.hh"
@@ -25,14 +23,15 @@
 #include "data_parser.hh"
 #include "data_scanner.hh"
 #include "elem_to_json.hh"
+#include "fmt/format.h"
 #include "formats/logfmt/logfmt.parser.hh"
+#include "hasher.hh"
 #include "libbase64.h"
 #include "mapbox/variant.hpp"
-#include "optional.hpp"
 #include "pcrepp/pcre2pp.hh"
+#include "pretty_printer.hh"
 #include "safe/safe.h"
-#include "scn/scn.h"
-#include "spookyhash/SpookyV2.h"
+#include "scn/scan.h"
 #include "sqlite-extension-func.hh"
 #include "text_anonymizer.hh"
 #include "view_curses.hh"
@@ -47,8 +46,6 @@
 #    include <curl/curl.h>
 #endif
 
-using namespace mapbox;
-
 enum class encode_algo {
     base64,
     hex,
@@ -57,7 +54,7 @@ enum class encode_algo {
 
 template<>
 struct from_sqlite<encode_algo> {
-    inline encode_algo operator()(int argc, sqlite3_value** val, int argi)
+    encode_algo operator()(int argc, sqlite3_value** val, int argi)
     {
         const char* algo_name = (const char*) sqlite3_value_text(val[argi]);
 
@@ -84,12 +81,12 @@ struct cache_entry {
         std::make_shared<column_namer>(column_namer::language::JSON)};
 };
 
-static cache_entry*
+cache_entry*
 find_re(string_fragment re)
 {
     using re_cache_t
         = std::unordered_map<string_fragment, cache_entry, frag_hasher>;
-    static thread_local re_cache_t cache;
+    thread_local re_cache_t cache;
 
     auto iter = cache.find(re);
     if (iter == cache.end()) {
@@ -117,7 +114,7 @@ find_re(string_fragment re)
     return &iter->second;
 }
 
-static bool
+bool
 regexp(string_fragment re, string_fragment str)
 {
     auto* reobj = find_re(re);
@@ -125,8 +122,9 @@ regexp(string_fragment re, string_fragment str)
     return reobj->re2->find_in(str).ignore_error().has_value();
 }
 
-static util::variant<int64_t, double, const char*, string_fragment, json_string>
-regexp_match(string_fragment re, string_fragment str)
+mapbox::util::
+    variant<int64_t, double, const char*, string_fragment, json_string>
+    regexp_match(string_fragment re, string_fragment str)
 {
     auto* reobj = find_re(re);
     auto& extractor = *reobj->re2;
@@ -147,9 +145,6 @@ regexp_match(string_fragment re, string_fragment str)
         throw std::runtime_error(err.get_message());
     }
 
-    yajlpp_gen gen;
-    yajl_gen_config(gen, yajl_gen_beautify, false);
-
     if (extractor.get_capture_count() == 1) {
         auto cap = md[1];
 
@@ -157,18 +152,24 @@ regexp_match(string_fragment re, string_fragment str)
             return static_cast<const char*>(nullptr);
         }
 
-        auto scan_int_res = scn::scan_value<int64_t>(cap->to_string_view());
-        if (scan_int_res && scan_int_res.empty()) {
-            return scan_int_res.value();
-        }
-
-        auto scan_float_res = scn::scan_value<double>(cap->to_string_view());
-        if (scan_float_res && scan_float_res.empty()) {
-            return scan_float_res.value();
+        auto scan_int_res = scn::scan_int<int64_t>(cap->to_string_view());
+        if (scan_int_res) {
+            if (scan_int_res->range().empty()) {
+                return scan_int_res->value();
+            }
+            auto scan_float_res
+                = scn::scan_value<double>(cap->to_string_view());
+            if (scan_float_res && scan_float_res->range().empty()) {
+                return scan_float_res->value();
+            }
         }
 
         return cap.value();
-    } else {
+    }
+
+    yajlpp_gen gen;
+    yajl_gen_config(gen, yajl_gen_beautify, false);
+    {
         yajlpp_map root_map(gen);
 
         for (size_t lpc = 0; lpc < extractor.get_capture_count(); lpc++) {
@@ -182,12 +183,12 @@ regexp_match(string_fragment re, string_fragment str)
             } else {
                 auto scan_int_res
                     = scn::scan_value<int64_t>(cap->to_string_view());
-                if (scan_int_res && scan_int_res.empty()) {
-                    yajl_gen_integer(gen, scan_int_res.value());
+                if (scan_int_res && scan_int_res->range().empty()) {
+                    yajl_gen_integer(gen, scan_int_res->value());
                 } else {
                     auto scan_float_res
                         = scn::scan_value<double>(cap->to_string_view());
-                    if (scan_float_res && scan_float_res.empty()) {
+                    if (scan_float_res && scan_float_res->range().empty()) {
                         yajl_gen_number(gen, cap->data(), cap->length());
                     } else {
                         yajl_gen_pstring(gen, cap->data(), cap->length());
@@ -272,7 +273,7 @@ logfmt2json(string_fragment line)
     return json_string(gen);
 }
 
-static std::string
+std::string
 regexp_replace(string_fragment str, string_fragment re, const char* repl)
 {
     auto* reobj = find_re(re);
@@ -280,28 +281,47 @@ regexp_replace(string_fragment str, string_fragment re, const char* repl)
     return reobj->re2->replace(str, repl);
 }
 
-static std::string
+std::optional<int64_t>
+sql_fuzzy_match(const char* pat, const char* str)
+{
+    if (pat == nullptr) {
+        return std::nullopt;
+    }
+
+    if (pat[0] == '\0') {
+        return 1;
+    }
+
+    int score = 0;
+
+    if (!fts::fuzzy_match(pat, str, score)) {
+        return std::nullopt;
+    }
+
+    return score;
+}
+
+string_fragment
 spooky_hash(const std::vector<const char*>& args)
 {
-    byte_array<2, uint64> hash;
-    SpookyHash context;
+    thread_local char hash_str_buf[hasher::STRING_SIZE];
 
-    context.Init(0, 0);
+    hasher context;
     for (const auto* const arg : args) {
         int64_t len = arg != nullptr ? strlen(arg) : 0;
 
-        context.Update(&len, sizeof(len));
+        context.update((const char*) &len, sizeof(len));
         if (arg == nullptr) {
             continue;
         }
-        context.Update(arg, len);
+        context.update(arg, len);
     }
-    context.Final(hash.out(0), hash.out(1));
+    context.to_string(hash_str_buf);
 
-    return hash.to_string();
+    return string_fragment::from_bytes(hash_str_buf, sizeof(hash_str_buf) - 1);
 }
 
-static void
+void
 sql_spooky_hash_step(sqlite3_context* context, int argc, sqlite3_value** argv)
 {
     auto* hasher
@@ -319,7 +339,7 @@ sql_spooky_hash_step(sqlite3_context* context, int argc, sqlite3_value** argv)
     }
 }
 
-static void
+void
 sql_spooky_hash_final(sqlite3_context* context)
 {
     auto* hasher
@@ -344,7 +364,7 @@ struct sparkline_context {
     std::vector<double> sc_values;
 };
 
-static void
+void
 sparkline_step(sqlite3_context* context, int argc, sqlite3_value** argv)
 {
     auto* sc = (sparkline_context*) sqlite3_aggregate_context(
@@ -367,7 +387,7 @@ sparkline_step(sqlite3_context* context, int argc, sqlite3_value** argv)
     }
 }
 
-static void
+void
 sparkline_final(sqlite3_context* context)
 {
     auto* sc = (sparkline_context*) sqlite3_aggregate_context(
@@ -394,7 +414,7 @@ sparkline_final(sqlite3_context* context)
     sc->~sparkline_context();
 }
 
-nonstd::optional<util::variant<blob_auto_buffer, sqlite3_int64, double>>
+std::optional<mapbox::util::variant<blob_auto_buffer, sqlite3_int64, double>>
 sql_gunzip(sqlite3_value* val)
 {
     switch (sqlite3_value_type(val)) {
@@ -423,10 +443,10 @@ sql_gunzip(sqlite3_value* val)
             return sqlite3_value_double(val);
     }
 
-    return nonstd::nullopt;
+    return std::nullopt;
 }
 
-nonstd::optional<blob_auto_buffer>
+std::optional<blob_auto_buffer>
 sql_gzip(sqlite3_value* val)
 {
     switch (sqlite3_value_type(val)) {
@@ -458,7 +478,7 @@ sql_gzip(sqlite3_value* val)
         }
     }
 
-    return nonstd::nullopt;
+    return std::nullopt;
 }
 
 #if defined(HAVE_LIBCURL)
@@ -475,7 +495,7 @@ get_curl_easy()
 }
 #endif
 
-static mapbox::util::variant<text_auto_buffer, auto_mem<char>, null_value_t>
+mapbox::util::variant<text_auto_buffer, auto_mem<char>, null_value_t>
 sql_encode(sqlite3_value* value, encode_algo algo)
 {
     switch (sqlite3_value_type(value)) {
@@ -555,7 +575,7 @@ sql_encode(sqlite3_value* value, encode_algo algo)
     ensure(false);
 }
 
-static mapbox::util::variant<blob_auto_buffer, auto_mem<char>>
+mapbox::util::variant<blob_auto_buffer, auto_mem<char>>
 sql_decode(string_fragment str, encode_algo algo)
 {
     switch (algo) {
@@ -572,15 +592,16 @@ sql_decode(string_fragment str, encode_algo algo)
             auto sv = str.to_string_view();
 
             while (!sv.empty()) {
-                int32_t value;
-                auto scan_res = scn::scan(sv, "{:2x}", value);
+                auto scan_res = scn::scan<int32_t>(sv, "{:2x}");
                 if (!scan_res) {
                     throw sqlite_func_error(
                         "invalid hex input at: {}",
                         std::distance(str.begin(), sv.begin()));
                 }
+                auto value = scan_res->value();
                 buf.push_back((char) (value & 0xff));
-                sv = scan_res.range_as_string_view();
+                sv = std::string_view{scan_res->range().begin(),
+                                      scan_res->range().size()};
             }
 
             return blob_auto_buffer{std::move(buf)};
@@ -605,7 +626,7 @@ sql_humanize_file_size(file_ssize_t value)
     return humanize::file_size(value, humanize::alignment::columnar);
 }
 
-static std::string
+std::string
 sql_anonymize(string_fragment frag)
 {
     static safe::Safe<lnav::text_anonymizer> ta;
@@ -620,7 +641,7 @@ const char* curl_url_strerror(CURLUcode error);
 }
 #endif
 
-static json_string
+json_string
 sql_parse_url(std::string url)
 {
     static auto* CURL_HANDLE = get_curl_easy();
@@ -631,9 +652,21 @@ sql_parse_url(std::string url)
     auto rc = curl_url_set(
         cu, CURLUPART_URL, url.c_str(), CURLU_NON_SUPPORT_SCHEME);
     if (rc != CURLUE_OK) {
-        throw lnav::console::user_message::error(
-            attr_line_t("invalid URL: ").append(lnav::roles::file(url)))
-            .with_reason(curl_url_strerror(rc));
+        auto_mem<char> url_part(curl_free);
+        yajlpp_gen gen;
+        yajl_gen_config(gen, yajl_gen_beautify, false);
+
+        {
+            yajlpp_map root(gen);
+            root.gen("error");
+            root.gen("invalid-url");
+            root.gen("url");
+            root.gen(url);
+            root.gen("reason");
+            root.gen(curl_url_strerror(rc));
+        }
+
+        return json_string(gen);
     }
 
     auto_mem<char> url_part(curl_free);
@@ -682,7 +715,18 @@ sql_parse_url(std::string url)
         root.gen("path");
         rc = curl_url_get(cu, CURLUPART_PATH, url_part.out(), CURLU_URLDECODE);
         if (rc == CURLUE_OK) {
-            root.gen(string_fragment::from_c_str(url_part.in()));
+            auto path_frag = string_fragment::from_c_str(url_part.in());
+            auto path_utf_res = is_utf8(path_frag);
+            if (path_utf_res.is_valid()) {
+                root.gen(path_frag);
+            } else {
+                rc = curl_url_get(cu, CURLUPART_PATH, url_part.out(), 0);
+                if (rc == CURLUE_OK) {
+                    root.gen(string_fragment::from_c_str(url_part.in()));
+                } else {
+                    root.gen();
+                }
+            }
         } else {
             root.gen();
         }
@@ -721,12 +765,24 @@ sql_parse_url(std::string url)
                 if (eq_index_opt) {
                     auto key = kv_pair_frag.sub_range(0, eq_index_opt.value());
                     auto val = kv_pair_frag.substr(eq_index_opt.value() + 1);
-                    auto key_str = key.to_string();
 
-                    if (seen_keys.count(key_str) == 0) {
-                        seen_keys.emplace(key_str);
-                        query_map.gen(key);
-                        query_map.gen(val);
+                    auto key_utf_res = is_utf8(key);
+                    auto val_utf_res = is_utf8(val);
+                    if (key_utf_res.is_valid()) {
+                        auto key_str = key.to_string();
+
+                        if (seen_keys.count(key_str) == 0) {
+                            seen_keys.emplace(key_str);
+                            query_map.gen(key);
+                            if (val_utf_res.is_valid()) {
+                                query_map.gen(val);
+                            } else {
+                                auto eq = strchr(kv_pair_encoded.data(), '=');
+                                query_map.gen(
+                                    string_fragment::from_c_str(eq + 1));
+                            }
+                        }
+                    } else {
                     }
                 } else {
                     auto val_str = split_res.first.to_string();
@@ -764,41 +820,48 @@ sql_parse_url(std::string url)
 }
 
 struct url_parts {
-    nonstd::optional<std::string> up_scheme;
-    nonstd::optional<std::string> up_username;
-    nonstd::optional<std::string> up_password;
-    nonstd::optional<std::string> up_host;
-    nonstd::optional<std::string> up_port;
-    nonstd::optional<std::string> up_path;
-    nonstd::optional<std::string> up_query;
-    std::map<std::string, nonstd::optional<std::string>> up_parameters;
-    nonstd::optional<std::string> up_fragment;
+    std::optional<std::string> up_scheme;
+    std::optional<std::string> up_username;
+    std::optional<std::string> up_password;
+    std::optional<std::string> up_host;
+    std::optional<std::string> up_port;
+    std::optional<std::string> up_path;
+    std::optional<std::string> up_query;
+    std::map<std::string, std::optional<std::string>> up_parameters;
+    std::optional<std::string> up_fragment;
 };
 
-static const json_path_container url_params_handlers = {
-    yajlpp::pattern_property_handler("(?<param>.*)")
-        .for_field(&url_parts::up_parameters),
-};
+const typed_json_path_container<url_parts>&
+get_url_parts_handlers()
+{
+    static const json_path_container url_params_handlers = {
+        yajlpp::pattern_property_handler("(?<param>.*)")
+            .for_field(&url_parts::up_parameters),
+    };
 
-static const typed_json_path_container<url_parts> url_parts_handlers = {
-    yajlpp::property_handler("scheme").for_field(&url_parts::up_scheme),
-    yajlpp::property_handler("username").for_field(&url_parts::up_username),
-    yajlpp::property_handler("password").for_field(&url_parts::up_password),
-    yajlpp::property_handler("host").for_field(&url_parts::up_host),
-    yajlpp::property_handler("port").for_field(&url_parts::up_port),
-    yajlpp::property_handler("path").for_field(&url_parts::up_path),
-    yajlpp::property_handler("query").for_field(&url_parts::up_query),
-    yajlpp::property_handler("parameters").with_children(url_params_handlers),
-    yajlpp::property_handler("fragment").for_field(&url_parts::up_fragment),
-};
+    static const typed_json_path_container<url_parts> retval = {
+        yajlpp::property_handler("scheme").for_field(&url_parts::up_scheme),
+        yajlpp::property_handler("username").for_field(&url_parts::up_username),
+        yajlpp::property_handler("password").for_field(&url_parts::up_password),
+        yajlpp::property_handler("host").for_field(&url_parts::up_host),
+        yajlpp::property_handler("port").for_field(&url_parts::up_port),
+        yajlpp::property_handler("path").for_field(&url_parts::up_path),
+        yajlpp::property_handler("query").for_field(&url_parts::up_query),
+        yajlpp::property_handler("parameters")
+            .with_children(url_params_handlers),
+        yajlpp::property_handler("fragment").for_field(&url_parts::up_fragment),
+    };
 
-static auto_mem<char>
+    return retval;
+}
+
+auto_mem<char>
 sql_unparse_url(string_fragment in)
 {
     static auto* CURL_HANDLE = get_curl_easy();
     static intern_string_t SRC = intern_string::lookup("arg");
 
-    auto parse_res = url_parts_handlers.parser_for(SRC).of(in);
+    auto parse_res = get_url_parts_handlers().parser_for(SRC).of(in);
     if (parse_res.isErr()) {
         throw parse_res.unwrapErr()[0];
     }
@@ -887,10 +950,23 @@ sql_humanize_id(string_fragment id)
     auto attrs = vc.attrs_for_ident(id.data(), id.length());
 
     return fmt::format(FMT_STRING("\x1b[38;5;{}m{}\x1b[0m"),
-                       attrs.ta_fg_color.value_or(COLOR_CYAN),
+                       // XXX attrs.ta_fg_color.value_or(COLOR_CYAN),
+                       (int8_t) ansi_color::cyan,
                        id);
 }
 
+static std::string
+sql_pretty_print(string_fragment in)
+{
+    data_scanner ds(in);
+    pretty_printer pp(&ds, {});
+    attr_line_t retval;
+
+    pp.append_to(retval);
+
+    return std::move(retval.get_string());
+}
+
 int
 string_extension_functions(struct FuncDef** basic_funcs,
                            struct FuncDefAgg** agg_funcs)
@@ -931,7 +1007,8 @@ string_extension_functions(struct FuncDef** basic_funcs,
                     "named properties 'num' and 'str'",
                     "SELECT regexp_match('(?<num>\\d+) (?<str>\\w+)', '123 "
                     "four')",
-                })),
+                }))
+            .with_result_subtype(),
 
         sqlite_func_adapter<decltype(&regexp_replace), regexp_replace>::builder(
             help_text("regexp_replace",
@@ -1045,7 +1122,8 @@ string_extension_functions(struct FuncDef** basic_funcs,
                 .with_example({
                     "To extract columnar data from a string",
                     "SELECT extract('1.0 abc 2.0')",
-                })),
+                }))
+            .with_result_subtype(),
 
         sqlite_func_adapter<decltype(&logfmt2json), logfmt2json>::builder(
             help_text("logfmt2json",
@@ -1057,7 +1135,8 @@ string_extension_functions(struct FuncDef** basic_funcs,
                 .with_example({
                     "To extract key/value pairs from a log message",
                     "SELECT logfmt2json('foo=1 bar=2 name=\"Rolo Tomassi\"')",
-                })),
+                }))
+            .with_result_subtype(),
 
         sqlite_func_adapter<
             decltype(static_cast<bool (*)(const char*, const char*)>(
@@ -1099,6 +1178,23 @@ string_extension_functions(struct FuncDef** basic_funcs,
                         "SELECT endswith('notbad.png', '.jpg')",
                     })),
 
+        sqlite_func_adapter<decltype(&sql_fuzzy_match), sql_fuzzy_match>::
+            builder(help_text(
+                        "fuzzy_match",
+                        "Perform a fuzzy match of a pattern against a "
+                        "string and return a score or NULL if the pattern was "
+                        "not matched")
+                        .sql_function()
+                        .with_parameter(help_text(
+                            "pattern", "The pattern to look for in the string"))
+                        .with_parameter(
+                            help_text("str", "The string to match against"))
+                        .with_tags({"string"})
+                        .with_example({
+                            "To match the pattern 'fo' against 'filter-out'",
+                            "SELECT fuzzy_match('fo', 'filter-out')",
+                        })),
+
         sqlite_func_adapter<decltype(&spooky_hash), spooky_hash>::builder(
             help_text("spooky_hash",
                       "Compute the hash value for the given arguments.")
@@ -1228,7 +1324,8 @@ string_extension_functions(struct FuncDef** basic_funcs,
                     "'https://alice@[fe80::14ff:4ee5:1215:2fb2]'",
                     "SELECT "
                     "parse_url('https://alice@[fe80::14ff:4ee5:1215:2fb2]')",
-                })),
+                }))
+            .with_result_subtype(),
 
         sqlite_func_adapter<decltype(&sql_unparse_url), sql_unparse_url>::
             builder(
@@ -1247,6 +1344,21 @@ string_extension_functions(struct FuncDef** basic_funcs,
                         "\"example.com\"}')",
                     })),
 
+        sqlite_func_adapter<decltype(&sql_pretty_print), sql_pretty_print>::
+            builder(
+                help_text("pretty_print", "Pretty-print the given string")
+                    .sql_function()
+                    .with_prql_path({"text", "pretty"})
+                    .with_parameter(help_text("str", "The string to format"))
+                    .with_tags({"string"})
+                    .with_example({
+                        "To pretty-print the string "
+                        "'{\"scheme\": \"https\", \"host\": \"example.com\"}'",
+                        "SELECT "
+                        "pretty_print('{\"scheme\": \"https\", \"host\": "
+                        "\"example.com\"}')",
+                    })),
+
         {nullptr},
     };
 
@@ -1254,6 +1366,7 @@ string_extension_functions(struct FuncDef** basic_funcs,
         {
             "group_spooky_hash",
             -1,
+            SQLITE_UTF8,
             0,
             sql_spooky_hash_step,
             sql_spooky_hash_final,
@@ -1273,6 +1386,7 @@ string_extension_functions(struct FuncDef** basic_funcs,
         {
             "sparkline",
             -1,
+            SQLITE_UTF8,
             0,
             sparkline_step,
             sparkline_final,
