@@ -29,6 +29,18 @@ if test -d meta-sessions/.lnav; then
     exit 1
 fi
 
+# tag was saved and search finds comment
+run_cap_test ${lnav_test} -n \
+    -c ":load-session" \
+    -c "/Hello, World" \
+    ${test_dir}/logfile_access_log.0
+
+# tag was saved and search finds tag
+run_cap_test ${lnav_test} -n \
+    -c ":load-session" \
+    -c "/foo" \
+    ${test_dir}/logfile_access_log.0
+
 # tag was saved and :write-to displays the comments/tags
 run_cap_test ${lnav_test} -n \
     -c ":load-session" \
@@ -134,3 +146,13 @@ run_cap_test ${lnav_test} -n \
     -c ";UPDATE access_log SET log_annotations = '{\"abc\": \"def\"}' WHERE log_line = 0" \
     -c ";SELECT log_line,log_annotations FROM access_log WHERE log_annotations IS NOT NULL" \
     ${test_dir}/logfile_access_log.0
+
+run_cap_test ${lnav_test} -n \
+    -c ":goto 0" \
+    -c ":annotate" \
+    ${test_dir}/logfile_postgres.1
+
+run_cap_test ${lnav_test} -n \
+    -c ":goto 0" \
+    -c ":annotate" \
+    ${test_dir}/logfile_postgres.2
